import math
import pandas as pd
from datetime import datetime
import bifrostapi
from bson.objectid import ObjectId
import bifrost_dashboard.components.global_vars as global_vars

pd.options.mode.chained_assignment = None


# Utils

def get_from_path(path_string, response):
    fields = path_string.split(".")
    for field in fields:
        response = response.get(field, None)
        if response is None:
            return response
    return response

# Main functions


def check_run_name(name):
    return bifrostapi.check_run_name(name)


def get_run_list():
    return bifrostapi.get_run_list()


def get_group_list(run_name=None):
    return bifrostapi.get_group_list(run_name)


def get_species_list(species_source, run_name=None):
    return bifrostapi.get_species_list(species_source, run_name)


def filter_name(species=None, group=None, qc_list=None, run_name=None):
    result = bifrostapi.filter({"name": 1, "sample_sheet.sample_name": 1},
                               run_name, species, group, qc_list)
    return list(result)


def filter_all(species=None, species_source=None, group=None,
               qc_list=None, run_names=None, sample_ids=None,
               sample_names=None,
               date_range=None,
               pagination=None,
               projection=None):
    return pd.io.json.json_normalize(bifrostapi.filter(
        species=species,
        species_source=species_source,
        group=group,
        qc_list=qc_list,
        run_names=run_names,
        sample_ids=sample_ids,
        sample_names=sample_names,
        date_range=date_range,
        pagination=pagination,
        projection=projection))


def get_assemblies_paths(samples):
    return bifrostapi.get_assemblies_paths(samples)


# For run_checker
def get_species_QC_values(ncbi_species):
    return bifrostapi.get_species_QC_values(ncbi_species)


def get_sample_QC_status(run):
    return bifrostapi.get_sample_QC_status(run)


def get_last_runs(run=None, n=12, runtype=None):
    return bifrostapi.get_last_runs(run, n, runtype)


def create_feedback_s_c(user, sample, value):
    if user == "":
        raise ValueError("Missing user value")
    component = get_component(name="user_feedback", version="1.0")
    if component is None:
        component = bifrostapi.save_component(global_vars.feedback_component)
    d = datetime.utcnow()
    now = d.replace(microsecond=math.floor(d.microsecond/1000)*1000)

    if value not in ("OK", "other", "core facility"):
        raise ValueError("Feedback value {} not valid".format(value))

    if value == "OK":
        status = "pass"
    else:
        status = "fail"

    summary = {
        "stamp": {
            "name": "user_feedback",
            "display_name": "User QC feedback",
            "value": value,
            "status": status,
            "reason": user,  # for now its just user
            "user": user
        }
    }

    s_c = {
        "component": {
            "_id": component["_id"],
            "name": component["name"]
        },
        "sample": {
            "_id": sample["_id"],
            "name": sample["name"]
        },
        "metadata": {
            "created_at": now,
            "updated_at": now,
        },
        "path": None,
        "status": "Success",
        "setup_date": now,
        "results": summary,
        "properties": {
            "summary": summary,
            "component": {
                "_id": component["_id"],
                "date": now,
            }
        }
        
    }
    return s_c


def add_user_feedback_to_properties(sample, s_c):
    """
    Adds/updates user feedback to sample properties
    """
    stamper = sample.get("properties", {}).get("stamper", {})
    stamper["component"] = s_c["properties"]["component"]
    summary = stamper.get("summary", {})

    old_value = summary.get("stamp").get("value")

    summary["stamp"] = s_c["properties"]["summary"]["stamp"]
    sample["properties"] = sample.get("properties", {})
    sample["properties"]["stamper"] = stamper
    return sample, old_value


def add_user_feedback(user, sample_id, value):
    """
    Submits user feedback to a sample
    """
    sample = bifrostapi.get_sample(ObjectId(sample_id))
    s_c = create_feedback_s_c(user, sample, value)
    sample, old_value = add_user_feedback_to_properties(sample, s_c)
    bifrostapi.save_sample_component(s_c)
    bifrostapi.save_sample(sample)
    return (sample["name"], old_value)


def add_batch_user_feedback_and_mail(feedback_pairs, user, email_config):
    """
    Main function called to send sample QC feedback 
    """
    email_pairs = []
    for sample_id, value in feedback_pairs:
        name, old_value = add_user_feedback(user, sample_id, value)
        runs = bifrostapi.get_sample_runs([ObjectId(sample_id)])
        run_names = [run["name"] for run in runs]
        if "fail:core facility" in (old_value, value):
            email_pairs.append(name, old_value, value, run_names)
    send_mail(email_pairs, user, email_config)


def get_component(name=None, version=None):
    return bifrostapi.get_component(name, version)


def get_run(run_name):
    return bifrostapi.get_run(run_name)


def send_mail(sample_info, user, email_config):
    """
    Sends email about sample feedback updates.
    """
    import smtplib
    from email.mime.multipart import MIMEMultipart
    from email.mime.text import MIMEText

    if len(sample_info) == 0:
        return

    short_samples = ",".join([pair[0] for pair in sample_info])[
        :60]  # Trimmed to 60 chars
    msg = MIMEMultipart("alternative")
    msg["From"] = email_config["email_from"]
    msg['Subject'] = 'Sample status change: "{}"'.format(short_samples)
    msg['To'] = email_config["email_to"]

    email_text = ('Automatic message:\nUser "{}" has changed the status of the following samples:\n'
                  '\nSample name                Old status            New status            Run name\n').format(
        user)
    email_html = ('<html><body>Automatic message:\nUser "{}" has changed the status of the following samples:\n'
                  '\n<pre>Sample name                Old status            New status            Run name\n').format(
        user)
    table = ""
    for pair in sample_info:
        table += "{:27s}{:22s}{:22s}{}\n".format(
            pair[0], pair[1], pair[2], ",".join(pair[3]))
    email_text += table
    email_html += table + "</pre></body></html>"
    msg.attach(MIMEText(email_text, 'plain'))
    msg.attach(MIMEText(email_html, 'html'))
    s = smtplib.SMTP('localhost')
    s.sendmail(msg["From"], msg["To"], msg.as_string())


def get_samples(sample_ids, projection=None):
    sample_ids = [ObjectId(id) for id in sample_ids]
    return bifrostapi.get_samples(sample_ids, projection=projection)


def get_comment(run_id):
    returned = bifrostapi.get_comment(run_id)
    if returned:
        return returned.get("Comments", None)
    else:
        return returned


def set_comment(run_id, comment):
    return bifrostapi.set_comment(run_id, comment)




import dash_html_components as html

import bifrost_dashboard.components.import_data as import_data

# def resequence_report():


def resequence_report(run_name=None):
    update_notice = (" SL: Supplying Lab, CF: Core Facility, CF(LF): "
                        "Core Facility (Library Fail). -: No data. "
                        "*: user submitted. "
                        "The table will update every 30s automatically.")
    if run_name is None:
        return None
    last_runs = import_data.get_last_runs(
        run_name, 12)  # Get last 12 runs
    last_runs_names = [run["name"] for run in last_runs]
    prev_runs_dict = import_data.get_sample_QC_status(last_runs)
    header = html.Tr([html.Th(html.Div(html.Strong("Sample")),
                              className="rotate")] +
                     list(map(lambda x: html.Th(html.Div(x),
                                                className="rotate"),
                              last_runs_names)))
    rows = [header]
    for name, p_runs in prev_runs_dict.items():
        if name == "Undetermined":
            continue
        row = []
        row.append(html.Td(name))

        sample_all_OKs = True
        print(name, p_runs)
        for index in range(len(last_runs)):
            if last_runs[index]["name"] in p_runs:
                className = "0"
                title = "Not Run"
                status = p_runs[last_runs[index]["name"]]
                print(status)
                if status.startswith("OK"):
                    className = "2"
                    title = "OK"
                elif status == "SL":
                    # sample_all_OKs = False # Disabled
                    className = "1"
                    title = "Supplying Lab"
                elif status.startswith("CF"):
                    # Wont be triggered by supplying because its after.
                    className = "-1"
                    sample_all_OKs = False
                    title = "Core Facility"
                else:
                    # to account for libray fails
                    sample_all_OKs = False
                row.append(
                    html.Td(status,
                            className="center status-" + className,
                            title=title))
            else:
                row.append(html.Td("-", className="center status-0"))

        if not sample_all_OKs:
            rows.append(html.Tr(row))
    table = html.Table(rows, className="unset-width-table", id="resequence-table")
    return html.Div([
        html.Div([
            html.H6("Resequence Report",
                    className="m-0 font-weight-bold text-primary")
        ], className="card-header py-3"),
        html.Div([
            html.P(update_notice),
            table
        ], className="card-body")
    ], className="card shadow mb-4")




import dash_html_components as html

import bifrost_dashboard.components.import_data as import_data


def link_to_files_div():
    return html.Div(id="link-to-files-div")


def link_to_files(sample_store):
    """Generates a script string """
    sample_ids = [str(s["_id"]) for s in sample_store]
    samples = import_data.filter_all(sample_ids=sample_ids,
                                     projection={"name": 1,
                                                 "properties.datafiles.summary.paired_reads": 1})
    samples["_id"] = samples["_id"].astype(str)
    samples = samples.set_index("_id")
    # Access samples by their id
    assemblies = import_data.get_assemblies_paths(sample_ids)
    reads_script = "mkdir samples\ncd samples\n"
    assemblies_script = "mkdir assemblies\ncd assemblies\n"
    reads_errors = []
    assemblies_errors = []

    samples_total = set()
    for _id, sample in samples.iterrows():
        try:
            reads_script += "#{}\nln -s {} .\nln -s {} .\n".format(
                sample["name"],
                sample["properties.datafiles.summary.paired_reads"][0],
                sample["properties.datafiles.summary.paired_reads"][1])
            samples_total.add((sample["name"], sample.name))
        except (KeyError, TypeError):
            reads_errors.append("Missing data for sample: {} - {}. In database:\n{}".format(
                sample.get("name", "None"),
                sample.name,
                sample.get(
                    "properties.datafiles.summary.paired_reads", "No data")
            ))
    if len(reads_errors):
        reads_html = [
            html.Div([
                html.Div([
                    html.H6("Read files",
                            className="m-0 font-weight-bold text-primary"),
                ], className="card-header py-3"),
                html.Div([
                    "A few errors occurred locating the read paths. If you need more info, " +
                    "please contact an admin.",
                    html.Pre("\n".join(reads_errors), className="error-pre"),
                    html.Pre(reads_script, className="folder-pre")
                ], className="card-body")
            ], className="card shadow")
        ]
    else:
        reads_html = [
            html.Div([
                html.Div([
                    html.H6("Read files",
                            className="m-0 font-weight-bold text-primary"),
                ], className="card-header py-3"),
                html.Div([
                    html.Pre(reads_script, className="folder-pre")
                ], className="card-body")
            ], className="card shadow")
        ]

    samples_assemblies = set()
    for assembly in assemblies:
        sample = samples.loc[str(assembly["sample"]["_id"]), :]
        assemblies_script += "#{}\nln -s {} {}\n".format(
            sample["name"],
            assembly["path"] + "/contigs.fasta",
            sample["name"] + "_contigs.fasta")
        samples_assemblies.add((sample["name"], sample.name))
    
    assemblies_missing = samples_total - samples_assemblies
    for sample in assemblies_missing:
        assemblies_errors.append("Missing data for sample: {} - {}.".format(sample[0], sample[1]))

    if len(assemblies_errors):
        assemblies_html = [
            html.Div([
                html.Div([
                    html.H6("Assemblies",
                            className="m-0 font-weight-bold text-primary"),
                ], className="card-header py-3"),
                html.Div([
                    "A few errors occurred locating the contigs. If you need more info, " +
                    "please contact an admin.",
                    html.Pre("\n".join(assemblies_errors),
                             className="error-pre"),
                    html.Pre(assemblies_script, className="folder-pre")
                ], className="card-body")
            ], className="card mb-4 shadow")
        ]
    else:
        assemblies_html = [
            html.Div([
                html.Div([
                    html.H6("Assemblies",
                            className="m-0 font-weight-bold text-primary"),
                ], className="card-header py-3"),
                html.Div([
                    html.Pre(assemblies_script, className="folder-pre")
                ], className="card-body")
            ], className="card mb-4 shadow")
        ]
    
    return html.Div([html.Div(assemblies_html), html.Div(reads_html)])



import dash_core_components as dcc
import dash_html_components as html
import plotly.graph_objs as go
from plotly import tools
import pandas as pd
import numpy as np
import bifrost_dashboard.components.import_data as import_data
import bifrost_dashboard.components.global_vars as global_vars


def aggregate_report():
    return html.Div([
        html.Div([
            html.Div([
                html.Div([
                    html.H6(
                        "QC parameters",
                        className="m-0 font-weight-bold text-primary"
                    ),
                ], className="card-header py-3"),
                html.Div([
                    html.Div(
                        [
                            html.Label("Plot species",
                                       htmlFor="plot-species"),
                            dcc.RadioItems(
                                options=[
                                    {"label": " Provided",
                                        "value": "provided"},
                                    {"label": " Detected",
                                        "value": "detected"},
                                ],
                                value="provided",
                                labelStyle={
                                    'margin': '0 0.5rem 0.5rem 0'},
                                id="plot-species-source"
                            ),
                            html.Div(dcc.Dropdown(
                                id="plot-species"
                            ), id="plot-species-div")
                        ],
                        className=""
                    ),

                    dcc.Graph(id="summary-plot"),
                ], className="card-body")
            ], className="card shadow")
        ], className="col-12 col-xl-8"),
        html.Div([
            html.Div([
                html.Div([
                    html.H6(
                        "MLST",
                        className="m-0 font-weight-bold text-primary"
                    ),
                ], className="card-header py-3"),
                html.Div([
                    dcc.Graph(id="mlst-plot")
                ], className="card-body")
            ], className="card shadow")
        ], className="col-12 col-xl-4")
    ], className="row")


# Callbacks

def aggregate_species_dropdown(sample_store, plot_species, selected_species):
    sample_ids = [s["_id"] for s in sample_store]

    plot_df = import_data.filter_all(
        sample_ids=sample_ids,
        projection={"properties": 1})

    species_col = "properties.species_detection.summary.detected_species"

    if plot_species == "provided":
        species_col = "properties.sample_info.summary.provided_species"
    elif plot_species == "detected":
        species_col = "properties.species_detection.summary.detected_species"

    if species_col not in plot_df:
        plot_df[species_col] = np.nan
    plot_df.loc[pd.isnull(plot_df[species_col]), species_col] = "Not classified"
    species_list = plot_df[species_col].unique()
    species_list = ["All species", ] + list(species_list)
    if selected_species == "Not classified" or selected_species is None or selected_species not in species_list:
        if species_list[0] == "Not classified" and len(species_list) > 1:
            selected_species = species_list[1]
        else:
            selected_species = species_list[0]
    species_list_options = [
        {
            "label": species,
            "value": species
        } for species in species_list]
    return [selected_species, species_list_options]


def update_aggregate_fig(selected_species, sample_store, plot_species_source):
    if len(sample_store) == 0:
        return {"data": []}, {"data": []}
    plot_values = global_vars.plot_values
    traces = []
    trace_ranges = []

    species_col = "properties.species_detection.summary.detected_species"
    if plot_species_source == "provided":
        species_col = "properties.sample_info.summary.provided_species"
    elif plot_species_source == "detected":
        species_col = "properties.species_detection.summary.detected_species"

    sample_ids = [s["_id"] for s in sample_store]
    plot_df = import_data.filter_all(
        sample_ids=sample_ids)

    if "properties.sample_info.summary.provided_species" not in plot_df:
        plot_df["properties.sample_info.summary.provided_species"] = np.nan

    plot_df.loc[pd.isnull(plot_df[species_col]),
                species_col] = "Not classified"

    # Some columns need to have some text extracted
    for col in global_vars.value_from_test:
        if col in plot_df.columns:
            new = plot_df[col].str.split(":", expand=True)
            loc = plot_df.columns.get_loc(col)
            plot_df.insert(loc + 1, col + ".value", new[2])
    sunburst_fig = generate_sunburst(plot_df)

    # Convert to string so the id can be shown in the plot.
    plot_df["_id"] = plot_df["_id"].astype(str)

    if species_col in plot_df.columns and (selected_species in plot_df[species_col].unique() or
                                           selected_species == "All species"):
        for plot_value in plot_values:
            plot_id = plot_value["id"]
            if selected_species == "All species":
                species_df = plot_df
            else:
                species_df = plot_df[plot_df[species_col] == selected_species]
            species_df[plot_id] = pd.to_numeric(
                species_df[plot_id], errors="coerce")
            if (plot_id in species_df.columns):
                data_range = plot_value["limits"][1] - plot_value["limits"][0]
                low_limit = min(
                    float(species_df[plot_id].min()), plot_value["limits"][0])
                if low_limit == float(species_df[plot_id].min()):
                    low_limit -= data_range * 0.1
                high_limit = max(
                    float(species_df[plot_id].max()), plot_value["limits"][1])
                if high_limit == float(species_df[plot_id].max()):
                    high_limit += data_range*0.1
                trace_ranges.append([low_limit, high_limit])
                traces.append(
                    go.Box(
                        x=species_df.loc[:, plot_id],
                        text=species_df["name"],
                        marker=dict(
                            size=4
                        ),

                        boxpoints="all",
                        jitter=0.3,
                        pointpos=-1.6,
                        selectedpoints=list(
                            range(len(species_df.index))),
                        name=plot_value["name"],
                        showlegend=False,
                        customdata=species_df["_id"]
                    )
                )
    fig = tools.make_subplots(rows=7, cols=1, print_grid=False)
    fig["layout"].update(
        hovermode="closest",
        title=selected_species,
        height=750,
        margin=go.layout.Margin(
            l=175,  # noqa
            r=50,
            b=25,
            t=50
        ),
    )
    try:
        fig.append_trace(traces[0], 1, 1)
        fig.append_trace(traces[1], 1, 1)
        fig.append_trace(traces[2], 2, 1)
        fig.append_trace(traces[3], 3, 1)
        fig.append_trace(traces[4], 4, 1)
        fig.append_trace(traces[5], 5, 1)
        fig.append_trace(traces[6], 6, 1)
        fig.append_trace(traces[7], 7, 1)

        fig["layout"]["xaxis"].update(range=trace_ranges[0])
        fig["layout"]["xaxis2"].update(range=trace_ranges[2])
        fig["layout"]["xaxis3"].update(range=trace_ranges[3])
        fig["layout"]["xaxis4"].update(range=trace_ranges[4])
        fig["layout"]["xaxis5"].update(range=trace_ranges[5])
        fig["layout"]["xaxis6"].update(range=trace_ranges[6])
        fig["layout"]["xaxis7"].update(range=trace_ranges[7])
    except IndexError:
        pass  # It won't draw al ranges/traces because they don't exist.

    fig["layout"]["yaxis"].update(domain=(0.78, 1))
    fig["layout"]["yaxis2"].update(domain=(0.655, 0.75))
    fig["layout"]["yaxis3"].update(domain=(0.53, 0.625))
    fig["layout"]["yaxis4"].update(domain=(0.415, 0.5))
    fig["layout"]["yaxis5"].update(domain=(0.28, 0.375))
    fig["layout"]["yaxis6"].update(domain=(0.155, 0.25))
    fig["layout"]["yaxis7"].update(domain=(0.03, 0.125))

    species_size = import_data.get_species_QC_values(selected_species)

    annotations = [
        {
            "x": species_size["min_length"],
            "y": 0,
            "xref": "x",
            "yref": "y",
            "text": "min",
            "arrowhead": 0,
            "ax": 0,
            "ay": 40
        },
        {
            "x": species_size["min_length"],
            "y": 1,
            "xref": "x",
            "yref": "y",
            "text": "min",
            "arrowhead": 0,
            "ax": 0,
            "ay": 40
        },
        {
            "x": species_size["max_length"],
            "y": 0,
            "xref": "x",
            "yref": "y",
            "text": "max",
            "arrowhead": 0,
            "ax": 0,
            "ay": 40
        },
        {
            "x": species_size["max_length"],
            "y": 1,
            "xref": "x",
            "yref": "y",
            "text": "max",
            "arrowhead": 0,
            "ax": 0,
            "ay": 40
        },
        {
            "x": 250000,
            "y": 0,
            "xref": "x2",
            "yref": "y2",
            "text": "max",
            "arrowhead": 0,
            "ax": 0,
            "ay": 40
        },
        {  # Cov
            "x": 10,
            "y": 0,
            "xref": "x3",
            "yref": "y3",
            "text": "fail",
            "arrowhead": 0,
            "ax": 0,
            "ay": 40
        },
        {  # Cov
            "x": 25,
            "y": 0,
            "xref": "x3",
            "yref": "y3",
            "text": "low",
            "arrowhead": 0,
            "ax": 0,
            "ay": 40
        },
        {  # Cov
            "x": 50,
            "y": 0,
            "xref": "x3",
            "yref": "y3",
            "text": "warn",
            "arrowhead": 0,
            "ax": 0,
            "ay": 40
        },
        {  # Num reads
            "x": 10000,
            "y": 0,
            "xref": "x5",
            "yref": "y5",
            "text": "min",
            "arrowhead": 0,
            "ax": 0,
            "ay": 40
        },
        {  # Main+uncl
            "x": 0.95,
            "y": 0,
            "xref": "x6",
            "yref": "y6",
            "text": "min",
            "arrowhead": 0,
            "ax": 0,
            "ay": 40
        },
        {  # Uncl
            "x": 0.2,
            "y": 0,
            "xref": "x7",
            "yref": "y7",
            "text": "max",
            "arrowhead": 0,
            "ax": 0,
            "ay": 40
        },
    ]

    fig["layout"].update(annotations=annotations)

    return fig, sunburst_fig

def generate_sunburst(plot_df):
    species_col = "properties.species_detection.summary.species"
    mlst_col = "properties.mlst.summary.strain"
    if not (species_col in plot_df.columns and mlst_col in plot_df.columns):
        return go.Figure()
    
    unique_species = plot_df[species_col].unique()
    labels = ["samples"]
    parents = [""]
    ids = ["samples"]
    values = [len(plot_df["_id"])]
    for row in plot_df.loc[plot_df[mlst_col].isnull(), mlst_col].index:
        plot_df.at[row, mlst_col] = [None]
    def joinmlst(l):
        return ', '.join(map(str, l))

    plot_df["properties.mlst.summary.strainstr"] = plot_df[mlst_col].apply(joinmlst)
    for species in unique_species:
        species_df = plot_df[plot_df[species_col] == species]
        ids.append(short_species(species))
        labels.append(short_species(species))
        parents.append("samples")
        values.append(len(species_df))
        if "properties.mlst.summary.strainstr" in species_df.columns:
            unique_mlst = species_df["properties.mlst.summary.strainstr"].unique(
            )
            for mlst in unique_mlst:
                ids.append("{} - {}".format(species, mlst))
                labels.append(mlst)
                parents.append(short_species(species))
                values.append(str(
                    len(species_df[species_df["properties.mlst.summary.strainstr"] == mlst])))
    trace = go.Sunburst(
        ids=ids,
        labels=labels,
        parents=parents,
        values=values,
        branchvalues="total",
        outsidetextfont={"size": 20, "color": "#377eb8"},
        marker={"line": {"width": 2}},
    )

    layout = go.Layout(
        margin=go.layout.Margin(t=0, l=0, r=0, b=0)
    )
    return go.Figure([trace], layout)


def short_species(species):
    if species is None or pd.isna(species):
        return None
    words = species.split(" ")
    if len(words) == 1:
        return species
    return "{}. {}".format(words[0][0], " ".join(words[1:]))



# -*- coding: utf-8 -*-
import os
import sys
import subprocess
import urllib.parse as urlparse

import dash
import dash_auth
import dash_core_components as dcc
import dash_html_components as html
import dash_table
import dash_bootstrap_components as dbc
import pandas as pd
import plotly.graph_objs as go
import plotly.figure_factory as ff
from dash.dependencies import Input, Output, State

import bifrost_dashboard.components.import_data as import_data
from datetime import datetime

import json
from bson import json_util

components_order = [
    "min_read_check", "whats_my_species", "assemblatron", "ssi_stamper",
    "cge_resfinder", "cge_mlst", "cge_plasmidfinder",
    "cge_virulencefinder", "ariba_resfinder", "ariba_mlst",
    "ariba_plasmidfinder", "ariba_virulencefinder", "sp_cdiff_fbi",
    "sp_ecoli_fbi", "sp_salm_fbi", "analyzer",
    "qcquickie", "testomatic"]


def pipeline_report(sample_data):
    update_notice = "The table will update every 30s automatically."

    table = dash_table.DataTable(
        id="pipeline-table", selected_rows=[],
        style_table={
            'overflowX': 'scroll',
            'width': '100%'
        },
        page_action='none',
        style_as_list_view=True,
        style_cell={
            'textAlign': 'center',
            "fontFamily": "Arial",
            "padding": "0px 10px",
            "fontSize": "0.7rem",
            "height": "25px"
        }
    )

    update_notice += (" Req.: requirements not met. Init.: initialised. "
                      "*: user submitted")
    rerun_columns = [
        {"id": "sample", "name": "sample"},
        {"id": "component", "name": "component"},
    ]

    return [
        # resequence_link,
        dbc.Row([
            dbc.Col([
                html.Div([
                    html.Div([
                        html.H6("Pipeline status",
                                className="m-0 font-weight-bold text-primary")
                    ], className="card-header py-3"),
                    html.Div([
                        html.P(update_notice),
                        table,
                        dcc.Interval(
                            id='table-interval',
                            interval=30*1000,  # in milliseconds
                            n_intervals=0
                        )
                    ], className="card-body")
                ], className="card shadow mb-4")
            ], width=9),
            dbc.Col([
                html.Div([
                    html.Div([
                        html.H6("Rerun components",
                                className="m-0 font-weight-bold text-primary")
                    ], className="card-header py-3"),
                    html.Div([
                        dbc.Alert(id="rerun-output",
                                  color="secondary",
                                  dismissable=True,
                                  is_open=False),
                        html.Label(html.Strong(
                            "Add all components for sample")),
                        dbc.InputGroup([
                            dcc.Dropdown(id="rerun-components",
                                         className="dropdown-group"),
                            dbc.InputGroupAddon(
                                dbc.Button("Add", id="rerun-add-components"),
                                addon_type="append",
                            ),
                        ]),
                        html.Label(html.Strong("Add component for all samples"),
                                   className="mt-3"),
                        dbc.InputGroup([
                            dcc.Dropdown(id="rerun-samples",
                                         className="dropdown-group"),
                            dbc.InputGroupAddon(
                                dbc.Button("Add", id="rerun-add-samples"),
                                addon_type="append",
                            ),
                        ]),
                        html.Label(html.Strong("Add all failed or init. components"),
                                   className="mt-3"),
                        dbc.Button("Add", id="rerun-add-failed", block=True),
                        html.Div([
                            html.H4("Selected sample components"),
                            dash_table.DataTable(
                                id="pipeline-rerun",
                                columns=rerun_columns,
                                row_deletable=True),
                            dbc.Button("Rerun selected sample components",
                                       id="rerun-button", className="mt-3", block=True,
                                       color="primary"),
                        ], className="mt-3")
                    ], className="card-body")
                ], className="card shadow mb-4")
            ], width=3
            )
        ])
    ]


def pipeline_report_data(sample_data):

    if len(sample_data) == 0:
        return [], [{"id": "Priority", "name": "Priority"},
                    {"id": "Sample", "name": "Sample"},
                    {"id": "QC status", "name": "QC status"}], [], [], []

    status_dict = {
        "Success": "OK",
        "Running": "Running",
        "initialized": "init.",
        "Failure": "Fail",
        "Requirements not met": "Req.",
        "queued to run": "queue",
    }
    samples = import_data.filter_all(
        sample_ids=[s["_id"] for s in sample_data],
        projection={"properties.stamper": 1, 'name': 1,
                    'sample_sheet.priority': 1,
                    "components": 1})

    s_c_components = []
    for _id, sample in samples.iterrows():
        try:
            for component in sample["components"]:
                s_c_components.append(component["name"])
        except TypeError:
            pass
    components_list = [
        comp for comp in components_order if comp in s_c_components]
    rows = []

    columns = [
        {"name": "Priority", "id": "priority"},
        {"name": "Sample", "id": "sample"},
        {"name": "QC status", "id": "qc_val"}
    ]

    rerun_form_components = []

    for comp in components_list:
        columns.append({"name": comp, "id": comp})
        # HERE
        rerun_form_components.append({"label": comp, "value": comp})

    # Conditional data colors
    style_data_conditional = [
        {
            "if": {
                "column_id": "qc_val",
                "filter_query": '{qc_val} contains "CF"'
            },
            "backgroundColor": "#ea6153"
        },
        {
            "if": {
                "column_id": "qc_val",
                "filter_query": '{qc_val} contains "CF(LF)"'
            },
            "backgroundColor": "#ea6153"
        },
        {
            "if": {
                "column_id": "qc_val",
                "filter_query": '{qc_val} contains "OK"'
            },
            "backgroundColor": "#27ae60"
        },
        {
            "if": {
                "column_id": "qc_val",
                "filter_query": '{qc_val} eq "SL"'
            },
            "backgroundColor": "#f1c40f"
        }
    ]
    for col in components_list:
        style_data_conditional.append({
            "if": {
                "column_id": col,
                "filter_query": '{{{}}} eq "Fail"'.format(col)
            },
            "backgroundColor": "#ea6153"
        })
        style_data_conditional.append({
            "if": {
                "column_id": col,
                "filter_query": '{{{}}} eq "OK"'.format(col)
            },
            "backgroundColor": "#3498db"
        })
        style_data_conditional.append({
            "if": {
                "column_id": col,
                "filter_query": '{{{}}} eq "Running"'.format(col)
            },
            "backgroundColor": "#f1c40f"
        })
        style_data_conditional.append({
            "if": {
                "column_id": col,
                "filter_query": '{{{}}} eq "Req."'.format(col)
            },
            "backgroundColor": "#d3d3d3",
            "color": "#525252"
        })

    rerun_form_samples = []
    for _id, sample in samples.iterrows():
        name = sample["name"]

        row = {}
        if name == "Undetermined":
            continue  # ignore this row
        row["sample"] = name
        row["_id"] = str(sample["_id"])

        rerun_form_samples.append({"label": name, "value": "{}:{}".format(
            _id, name)})

        priority = str(sample.get("sample_sheet.priority", "")).lower()
        # prio_display = " " Emoji not supported
        # if priority == "high":
        #     prio_display = "High"
        # else:
        #     prio_display = "Low"
        row["priority"] = priority
        qc_val = sample.get("properties.stamper.summary.stamp.value", "N/A")
        if pd.isna(qc_val):
            qc_val = "N/A"

        expert_check = False
        expert_stamp = sample.get('stamps.supplying_lab_check.value')
        if expert_stamp is not None and not pd.isna(expert_stamp):
            qc_val = sample.get('stamps.supplying_lab_check.value')
            expert_check = True

        statusname = ""
        if qc_val == "supplying lab":
            qc_val = "SL"
            statusname = "status-1"
        elif qc_val == "N/A":
            statusname = "status--2"
        elif (qc_val == "core facility"):
            statusname = "status--1"
            qc_val = "CF"
        elif qc_val == "OK":
            statusname = "status-2"
            qc_val = "OK"

        if expert_check:
            try:
                qc_val += "*"
            except:
                print(sample)

        row["qc_val"] = qc_val
        try:
            for component in sample["components"]:
                row[component["name"]] = status_dict[component["status"]]
                row["{}_id".format(component["name"])] = str(component["_id"])
        except TypeError:
            pass
        except KeyError:
            print(sample["_id"])
            print(sample["components"])
        rows.append(row)

    def sort_name(e):
        return e["sample"]
    rows.sort(key=sort_name)
    return rows, columns, style_data_conditional, rerun_form_components, rerun_form_samples


def rerun_components_button(button, table_data, rerun_conf):
    if button == 0:
        return "", False
    out = []
    to_rerun = {}
    for row in table_data:
        sample_rerun = to_rerun.get(row["sample_id"], [])
        sample_rerun.append((row["component"], row["component_id"]))
        to_rerun[row["sample_id"]] = sample_rerun

    sample_dbs = import_data.get_samples(sample_ids=to_rerun.keys(),
                                         projection={"name": 1, "path": 1,
                                                     "properties.datafiles.summary.paired_reads": 1})
    samples_by_id = {str(s["_id"]): s for s in sample_dbs}

    bifrost_components_dir = os.path.join(
        rerun_conf["bifrost_dir"], "components/")

    for sample, components in to_rerun.items():
        sample_db = samples_by_id[sample]
        sample_id = str(sample_db["_id"])
        sample_name = sample_db["name"]
        sample_path = sample_db["path"]
        reads = sample_db["properties"]["datafiles"]["summary"]["paired_reads"]
        sample_command = ""
        for component_pair in components:
            component = component_pair[0]
            component_id = component_pair[1]
            component_path = os.path.join(bifrost_components_dir,
                                          component, "pipeline.smk")
            db_key = os.environ.get('BIFROST_DB_KEY')
            command = (r'source activate env_snakemake; '
                       r'export BIFROST_DB_KEY={}; '
                       r'if [ -d \"{}\" ]; then rm -r {}; fi; ').format(
                db_key, component, component)
            sing_args = "-B " + reads[0] + "," + reads[1] + "," + \
                sample_path
            snakemake_command = (r'snakemake --use-singularity  --singularity-args \"{}\" '
                                 r'--singularity-prefix \"{}\" --restart-times 2 '
                                 r"--cores 4 -s {} "
                                 r"--config sample_id={} component_id={} {}; ")
            # unlock first
            command += snakemake_command.format(
                sing_args, rerun_conf["singularity_prefix"], component_path, sample_id, component_id, "--unlock")
            command += snakemake_command.format(
                sing_args, rerun_conf["singularity_prefix"], component_path, sample_id, component_id, "")
            sample_command += command

        if rerun_conf["grid"] == "slurm":
            process_command = ('sbatch --mem={memory}G -p {priority} -c {threads} '
                               '-t {walltime} -J "bifrost_{sample_name}" --wrap'
                               ' "{command}"').format(
                **rerun_conf,
                sample_name=sample_name,
                command=sample_command)
            process = subprocess.Popen(
                process_command,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                shell=True,
                env=os.environ,
                cwd=sample_path)
            process_out, process_err = process.communicate()
            out.append((sample_name, process_out, process_err))
        elif rerun_conf["grid"] == "torque":

            if "advres" in rerun_conf:
                advres = ",advres={}".format(
                    rerun_conf["advres"])
            else:
                advres = ''
            script_path = os.path.join(sample_path, "manual_rerun.sh")
            with open(script_path, "w") as script:
                command += ("#PBS -V -d . -w . -l mem={memory}gb,nodes=1:"
                            "ppn={threads},walltime={walltime}{advres} -N "
                            "'bifrost_{sample_name}' -W group_list={group}"
                            " -A {group} \n").format(**rerun_conf,
                                                     sample_name=sample_name)
                script.write(command)
            process = subprocess.Popen('qsub {}'.format(script_path),
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.STDOUT,
                                       shell=True,
                                       env=os.environ,
                                       cwd=sample_path)
            process_out, process_err = process.communicate()
            out.append((sample_name, process_out, process_err))
        elif rerun_conf["grid"] == "slurm.mock":
            print(('sbatch --mem={memory}G -p {priority} -c {threads} '
                   '-t {walltime} -J "bifrost_{sample_name}" --wrap'
                   ' "{command}"').format(
                **rerun_conf,
                sample_name=sample_name,
                command=sample_command))

    message = "Jobs sent to the server:\n"
    message += "\n".join(["{}: out: {} | err: {}".format(*el)
                          for el in out])
    return message, True


def update_rerun_table(active, table_data, n_click_comp, n_click_samp,
                       n_click_fail, columns, prev_data, rerun_comp,
                       rerun_samp):
    # default values
    if prev_data is None:
        prev_data = []

    if columns:
        columns = columns[3:]
        columns = [x for x in columns if not x["id"].endswith("_id")]

    # Get context to know which button was triggered.
    ctx = dash.callback_context

    if not ctx.triggered:
        triggered_id = None
    else:
        triggered_id = ctx.triggered[0]['prop_id']

    # Nothing triggered it, return empty table if init call or prev data.
    if active is None and triggered_id is None:
        return prev_data

    if triggered_id == "pipeline-table.active_cell":
        col = active["column_id"]
        sample = table_data[active["row"]]["sample"]
        sample_id = table_data[active["row"]]["_id"]
        component_id = table_data[active["row"]].get("{}_id".format(col), None)
        if col in components_order:
            new_rows = [{"sample": sample,
                         "component": col,
                         "sample_id": sample_id,
                         "component_id": component_id}]
        else:
            new_rows = []
    elif triggered_id == "rerun-add-components.n_clicks":
        sample_id, sample = rerun_comp.split(":")
        row = table_data[table_data._id == sample_id]
        new_rows = [{"sample": sample,
                     "component": comp["id"],
                     "sample_id": sample_id,
                     "component_id": row["{}_id".format(comp["id"])]} for comp in columns]
    elif triggered_id == "rerun-add-samples.n_clicks":
        new_rows = []
        for row in table_data:
            new_rows.append({"sample": row["sample"],
                             "component": rerun_samp,
                             "sample_id": row["_id"],
                             "component_id": row["{}_id".format(rerun_samp)]})
    elif triggered_id == "rerun-add-failed.n_clicks":
        new_rows = []
        for row in table_data:
            for col in columns:
                col = col["id"]
                if row[col] == "Fail" or row[col] == "init.":
                    new_rows.append({"sample": row["sample"],
                                     "component": col,
                                     "sample_id": row["_id"],
                                     "component_id": row["{}_id".format(col)]})
    else:
        new_rows = []

    for new_row in new_rows:
        if new_row not in prev_data:
            prev_data = prev_data + [new_row]

    return prev_data



import os
import dash_html_components as html
import dash_core_components as dcc
import dash_bootstrap_components as dbc


def html_qc_expert_form():
    if ("REPORTER_ADMIN" in os.environ and os.environ["REPORTER_ADMIN"] == "True"):
        return html.Div([
            dbc.Row([
                dbc.Col(html.H5("Submit Supplying Lab Feedback:",
                                className="nomargin"), width=6),
                dbc.Col(dcc.Input(id="qc-user-1", placeholder="Your initials (e.g. MBAS)",
                                  value="", type="text"), width=3),
                html.Div(id="placeholder0", style={"display": "none"}),
                dbc.Col([
                    html.Button("Submit QC values",
                                id="feedback-button"),
                    html.Div(id="qc-feedback", className="qc-feedback")], width=3)
            ]),
            html.Div([
                html.Details([
                    html.Summary("How it works (click to show)"),
                    html.P("""If you want to send a sample to resequence,
                    go to the sample card and select "Resequence" in 
                    the Suplying Lab Feedback section. The admin will get
                    an email with the change. You can also "Accept" samples
                    with the "Supplying lab" warning and their data will
                    be used in the future to adjust QC thresholds.""")
                ])
            ], className="mb-1")

        ])
        
    else:
        return None


def sample_radio_feedback(sample, n_sample):
    if ("REPORTER_ADMIN" in os.environ and os.environ["REPORTER_ADMIN"] == "True"):
        return (html.Div([
            html.Label("Supplying Lab Feedback:", className="mr-4",
                       style={'display': 'inline-block'}),
            dcc.RadioItems(
                options=[
                    {'label': ' Accept', 'value': 'OK_{}'.format(
                                        sample["_id"])},
                    {'label': ' Resequence', 'value': 'CF_{}'.format(
                        sample["_id"])},
                    {'label': ' Other', 'value': 'O_{}'.format(
                        sample["_id"])},
                    {'label': ' No action', 'value': 'noaction'}
                ],
                value='noaction',
                id="sample-radio-{}".format(n_sample),
                labelStyle={
                    'margin': '0 0.5rem 0.5rem 0'},
                style={'display': 'inline-block'}),
        ]))
    else:
        return None



from setuptools import setup, find_packages

setup(
    name='bifrost-dashboard',
    version='0.0.2',
    description='Dashboard for displaying bifrost information',
    url='https://github.com/ssi-dk/bifrost',
    author="Martin Basterrechea",
    author_email="mbas@ssi.dk",
    packages=find_packages(),
    package_data={
        "bifrost_dashboard": ['data/assets/*.css', 'data/assets/*.js']
    },
    include_package_data=True,
    install_requires=[
        # 'pymongo',
        'pandas',
        'dash',
        'bifrostapi',
        'Flask-Caching',
        'dash-auth',
        'requests',
        'dash-bootstrap-components',
        'pyyaml',
        #'uwsgi'  # doesnt work for me
        ],
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'bifrost_dashboard_dev=bifrost_dashboard.reporter:main'
        ]
    },
    scripts=['bin/bifrost_dashboard_start']
)



import os
import dash_auth
import bifrost_dashboard.reporter
import yaml
config = yaml.safe_load(open(os.environ["BIFROST_DASH_CONFIG"]))

bifrost_dashboard.reporter.ADMIN = True

app = bifrost_dashboard.reporter.app

dash_auth.BasicAuth(
    app,
    config['USERNAME_PASSWORD']
)

server = app.server



import pandas as pd
import dash_html_components as html

def html_table(data, header=None, **kwargs):
    rows = []
    if header is not None:
        thead = html.Thead(html.Tr([html.Th(h) for h in header]))
    else:
        thead = None
    for data_row in data:
        if isinstance(data_row, list):
            rows.append(
                html.Tr([html.Td(data_cell, className="cell") for data_cell in data_row], className="trow"))
        else:
            rows.append(html.Tr([html.Td(data_cell, className="cell") for data_cell in data_row["list"]],
                               className=data_row["className"] + " trow"))
    return html.Table([thead, html.Tbody(rows)], className="bifrost-table", **kwargs)


def html_td_percentage(value, color):
    string = str(round(float(value) * 100, 2)) + "%"
    if pd.isna(value):
        fill = "0%"
    else:
        fill = string
    return html.Td(
        html.Div([
            html.Span(string, className="val"),
            html.Span(
                className="bar",
                style={"backgroundColor": color, "width": fill}
            )
        ], className="wrapper"
        ), className="data-colored cell")



import math

def assemblatron_diff(res):
    if "properties.denovo_assembly.summary.bin_length_at_1x" in res and "properties.denovo_assembly.summary.bin_length_at_10x" in res:
        res["properties.denovo_assembly.summary.bin_length_1x_10x_diff"] = res["properties.denovo_assembly.summary.bin_length_at_1x"] - \
            res["properties.denovo_assembly.summary.bin_length_at_10x"]
    else:
        res["properties.denovo_assembly.summary.bin_length_1x_10x_diff"] = math.nan
    return res

def assemblatron_contig_diff(res):
    if "properties.denovo_assembly.summary.bin_contigs_at_1x" in res and "properties.denovo_assembly.summary.bin_contigs_at_10x" in res:
        res["properties.denovo_assembly.summary.bin_contigs_1x_10x_diff"] = res["properties.denovo_assembly.summary.bin_contigs_at_1x"] - \
            res["properties.denovo_assembly.summary.bin_contigs_at_10x"]
    else:
        res["properties.denovo_assembly.summary.bin_contigs_1x_10x_diff"] = math.nan
    return res

FUNCS = [assemblatron_diff, assemblatron_contig_diff]


COLUMNS = [
    {
        "name": "Run",
        "id": "sample_sheet.run_name"
    },
    {
        "name": "Name",
        "id": "name"
    },
    {
        "name": "Supplying_lab_feedback",
        "id": "stamps.supplying_lab_check.value",
    },
    {
        "name": "QC_action",
        "id": "properties.stamper.summary.stamp.value",
    },
    {
        "name": "Comments",
        "id": "sample_sheet.Comments"
    },
    {
        "name": "Supplying_lab",
        "id": "sample_sheet.group"
    },
    {
        "name": "Provided_Species",
        "id": "properties.sample_info.summary.provided_species"
    },
    {
        "name": "Detected Species",
        "id": "properties.species_detection.summary.detected_species"
    },
    {
        "name": "Genome_size_1x",
        "id": "properties.denovo_assembly.summary.bin_length_at_1x"
    },
    {
        "name": "Genome_size_10x",
        "id": "properties.denovo_assembly.summary.bin_length_at_10x"
    },
    {
        "name": "G_size_diff_1x_10",
        "id": "properties.stamper.summary.test__denovo_assembly__genome_size_difference_1x_10x.value"
    },
    {
        "name": "Avg_coverage",
        "id": "properties.denovo_assembly.summary.bin_coverage_at_1x"
    },
    {
        "name": "Num_contigs",
        "id": "properties.denovo_assembly.summary.bin_contigs_at_1x"
    },
    {
        "name": "Ambiguous_sites",
        "id": "properties.denovo_assembly.summary.snp_filter_10x_10%"
    },
    {
        "name": "Num_reads",
        "id": "properties.denovo_assembly.summary.filtered_reads_num"
    },
    {
        "name": "mlst_type",
        "id": "properties.mlst.summary.strain"
    },
    {
        "name": "Main_sp_plus_uncl",
        "id": "properties.stamper.summary.test__species_detection__main_species_level.value"
    },
    {
        "name": "Unclassified_reads",
        "id": "properties.species_detection.summary.percent_unclassified"
    },
    {
        "name": "File path",
        "id": "path"
    },
    {
        "name": "DB_ID",
        "id": "_id"
    },
    {
        "name": "Failed_tests",
        "id": "ssi_stamper_failed_tests"
    }
]

plot_values = [
    {
        "name": "Genome_size_1x",
        "id": "properties.denovo_assembly.summary.bin_length_at_1x",
        "limits": [1500000, 6000000]
    },
    {
        "name": "Genome_size_10x",
        "id": "properties.denovo_assembly.summary.bin_length_at_10x",
        "limits": [1500000, 6000000],
        "xaxis": "x"
    },
    {
        "name": "G_size_difference_1x_10",
        "id": "properties.stamper.summary.test__denovo_assembly__genome_size_difference_1x_10x.value",
        "limits": [0, 260000]
    },
    {
        "name": "Avg_coverage",
        "id": "properties.denovo_assembly.summary.bin_coverage_at_1x",
        "limits": [0, 200]
    },
    {
        "name": "Contig_num_1x",
        "id": "properties.denovo_assembly.summary.bin_contigs_at_1x",
        "limits": [0, 700]
    },
    {
        "name": "Num_reads",
        "id": "properties.denovo_assembly.summary.filtered_reads_num",
        "limits": [1000, 8000000]
    },
    {
        "name": "Main_sp_plus_unclassified",
        "id": "properties.stamper.summary.test__species_detection__main_species_level.value",
        "limits": [0.75, 1]
    },
    {
        "name": "Unclassified_reads",
        "id": "properties.species_detection.summary.percent_unclassified",
        "limits": [0, 0.25]
    }
]

value_from_test = ["properties.stamper.summary.test__denovo_assembly__genome_size_difference_1x_10x",
                   "properties.stamper.summary.test__species_detection__main_species_level"]

ROUND_COLUMNS = ["properties.species_detection.summary.percent_unclassified",
                 "properties.denovo_assembly.summary.bin_coverage_at_1x"]

expected_results = ["resistance", "mlst", "plasmid",
                    "virulence"]  # expected categories in sample report

feedback_component = {
    "name": "user_feedback",
    "version": "1.0",
    "details": {
        "category": "stamper",
        "schema_version": 2,
        "target": "sample",
        "type": "stamper",
        "recommendation": "optional",
        "description": "This is generated by the web reporter when a user submits feedback.",
    },
    "dockerfile": None,
    "options": {},
    "resources": {},
    "requirements": {},
    "db_values_changes": {
        "sample": {
            "properties": {
                "stamper": {
                    "summary": None,
                    "component": {
                            "_id": None,
                            "date": None
                    }
                }
            }
        },
        "sample_component": {
            "summary": {
                "stamp": {
                    "name": None,
                    "value": None,
                    "display_name": None,
                    "date": None,
                    "status": None,
                    "reason": None
                }
            }
        }
    }
}
